from .compound import (
    CatalyzeConvert,
    Equilibration,
    ReversibleSynthesis,
)
from .enzymatic import (
    MichaelisMenten,
    MichaelisMentenEqApprox,
    MichaelisMentenQuasiSSAprox,
)
from .single import (
    AutoCreation,
    Conversion,
    Creation,
    Destruction,
    Dissociation,
    Synthesis,
)

__all__ = [
    "ReversibleSynthesis",
    "Equilibration",
    "CatalyzeConvert",
    "MichaelisMenten",
    "MichaelisMentenEqApprox",
    "MichaelisMentenQuasiSSAprox",
    "Creation",
    "AutoCreation",
    "Destruction",
    "Conversion",
    "Synthesis",
    "Dissociation",
]



import numpy as np
from pytest import mark

from .. import (
    Simulator,
    System,
)
from ..reactions import compound, enzymatic, single

reactions = []
for mod in (single, compound, enzymatic):
    for name in dir(mod):
        value = getattr(mod, name)
        if isinstance(value, type(System)) and value != System:
            reactions.append(value)


@mark.parametrize("reaction", reactions)
def test_reactions(reaction):
    model = reaction(**dict.fromkeys(reaction._required, 1))
    sim = Simulator(model)
    sim.solve(save_at=np.linspace(0, 1, 10))



import functools
import keyword
from collections import ChainMap

import libsbml
from libsbml import ASTNode
from symbolite import Real
from symbolite.abstract import real
from symbolite.core import as_function

from .symbol import MathMLSpecialSymbol, MathMLSymbol


def as_symbol(node: libsbml.ASTNode):
    return MathMLSymbol(node.getName())


def as_special_symbol(node: libsbml.ASTNode):
    return MathMLSpecialSymbol(node.getName())


def get_value(cast_to: type):
    def getter(x: ASTNode):
        value = x.getValue()
        return cast_to(value)

    return getter


def minus(*args):
    if len(args) == 1:
        return real.neg(*args)
    elif len(args) == 2:
        return real.sub(*args)
    else:
        raise TypeError("Received AST_MINUS with more than 2 children.")


def root(*args):
    match args:
        case (arg,) | (2, arg):
            return real.sqrt(arg)
        case (exp, arg):
            return real.pow(arg, real.truediv(1, exp))
        case _:
            raise TypeError(f"unexpected number of arguments: {len(args)}")


def log(*args):
    match args:
        case (10, x) | (x,):
            return real.log10(x)
        case _:
            raise TypeError(f"unexpected arguments for log: {args}")


def reduced(func):
    def impl(*args):
        return functools.reduce(func, args)

    return impl


mapper = {
    libsbml.AST_PLUS: real.add,
    libsbml.AST_MINUS: minus,
    libsbml.AST_TIMES: real.mul,
    libsbml.AST_DIVIDE: real.truediv,
    libsbml.AST_POWER: real.pow,
    libsbml.AST_INTEGER: get_value(int),
    libsbml.AST_REAL: get_value(float),
    libsbml.AST_REAL_E: get_value(float),
    libsbml.AST_RATIONAL: "AST_RATIONAL",
    libsbml.AST_NAME: as_symbol,
    libsbml.AST_NAME_AVOGADRO: "AST_NAME_AVOGADRO",
    libsbml.AST_NAME_TIME: as_special_symbol,
    libsbml.AST_CONSTANT_E: real.e,
    libsbml.AST_CONSTANT_FALSE: False,
    libsbml.AST_CONSTANT_PI: real.pi,
    libsbml.AST_CONSTANT_TRUE: True,
    libsbml.AST_LAMBDA: "AST_LAMBDA",
    libsbml.AST_FUNCTION: "AST_FUNCTION",
    libsbml.AST_FUNCTION_ABS: real.abs,
    libsbml.AST_FUNCTION_ARCCOS: real.acos,
    libsbml.AST_FUNCTION_ARCCOSH: real.acosh,
    libsbml.AST_FUNCTION_ARCCOT: "AST_FUNCTION_ARCCOT",
    libsbml.AST_FUNCTION_ARCCOTH: "AST_FUNCTION_ARCCOTH",
    libsbml.AST_FUNCTION_ARCCSC: "AST_FUNCTION_ARCCSC",
    libsbml.AST_FUNCTION_ARCCSCH: "AST_FUNCTION_ARCCSCH",
    libsbml.AST_FUNCTION_ARCSEC: "AST_FUNCTION_ARCSEC",
    libsbml.AST_FUNCTION_ARCSECH: "AST_FUNCTION_ARCSECH",
    libsbml.AST_FUNCTION_ARCSIN: real.asin,
    libsbml.AST_FUNCTION_ARCSINH: real.asinh,
    libsbml.AST_FUNCTION_ARCTAN: real.atan,
    libsbml.AST_FUNCTION_ARCTANH: real.atanh,
    libsbml.AST_FUNCTION_CEILING: real.ceil,
    libsbml.AST_FUNCTION_COS: real.cos,
    libsbml.AST_FUNCTION_COSH: real.cosh,
    libsbml.AST_FUNCTION_COT: "AST_FUNCTION_COT",
    libsbml.AST_FUNCTION_COTH: "AST_FUNCTION_COTH",
    libsbml.AST_FUNCTION_CSC: "AST_FUNCTION_CSC",
    libsbml.AST_FUNCTION_CSCH: "AST_FUNCTION_CSCH",
    libsbml.AST_FUNCTION_DELAY: "AST_FUNCTION_DELAY",
    libsbml.AST_FUNCTION_EXP: real.exp,
    libsbml.AST_FUNCTION_FACTORIAL: real.factorial,
    libsbml.AST_FUNCTION_FLOOR: real.floor,
    libsbml.AST_FUNCTION_LN: real.log,
    libsbml.AST_FUNCTION_LOG: log,
    libsbml.AST_FUNCTION_PIECEWISE: "AST_FUNCTION_PIECEWISE",
    libsbml.AST_FUNCTION_POWER: real.pow,
    libsbml.AST_FUNCTION_ROOT: root,
    libsbml.AST_FUNCTION_SEC: "AST_FUNCTION_SEC",
    libsbml.AST_FUNCTION_SECH: "AST_FUNCTION_SECH",
    libsbml.AST_FUNCTION_SIN: real.sin,
    libsbml.AST_FUNCTION_SINH: real.sinh,
    libsbml.AST_FUNCTION_TAN: real.tan,
    libsbml.AST_FUNCTION_TANH: real.tanh,
    libsbml.AST_LOGICAL_AND: reduced(real.and_),
    libsbml.AST_LOGICAL_NOT: real.invert,
    libsbml.AST_LOGICAL_OR: reduced(real.or_),
    libsbml.AST_LOGICAL_XOR: real.xor,
    libsbml.AST_RELATIONAL_EQ: reduced(real.eq),
    libsbml.AST_RELATIONAL_GEQ: real.ge,
    libsbml.AST_RELATIONAL_GT: real.gt,
    libsbml.AST_RELATIONAL_LEQ: real.le,
    libsbml.AST_RELATIONAL_LT: real.lt,
    libsbml.AST_RELATIONAL_NEQ: reduced(real.ne),
    libsbml.AST_END_OF_CORE: "AST_END_OF_CORE",
    libsbml.AST_FUNCTION_MAX: "AST_FUNCTION_MAX",
    libsbml.AST_FUNCTION_MIN: "AST_FUNCTION_MIN",
    libsbml.AST_FUNCTION_QUOTIENT: "AST_FUNCTION_QUOTIENT",
    libsbml.AST_FUNCTION_RATE_OF: "AST_FUNCTION_RATE_OF",
    libsbml.AST_FUNCTION_REM: "AST_FUNCTION_REM",
    libsbml.AST_LOGICAL_IMPLIES: "AST_LOGICAL_IMPLIES",
    libsbml.AST_CSYMBOL_FUNCTION: "AST_CSYMBOL_FUNCTION",
    libsbml.AST_DISTRIB_FUNCTION_NORMAL: "AST_DISTRIB_FUNCTION_NORMAL",
    libsbml.AST_DISTRIB_FUNCTION_UNIFORM: "AST_DISTRIB_FUNCTION_UNIFORM",
    libsbml.AST_DISTRIB_FUNCTION_BERNOULLI: "AST_DISTRIB_FUNCTION_BERNOULLI",
    libsbml.AST_DISTRIB_FUNCTION_BINOMIAL: "AST_DISTRIB_FUNCTION_BINOMIAL",
    libsbml.AST_DISTRIB_FUNCTION_CAUCHY: "AST_DISTRIB_FUNCTION_CAUCHY",
    libsbml.AST_DISTRIB_FUNCTION_CHISQUARE: "AST_DISTRIB_FUNCTION_CHISQUARE",
    libsbml.AST_DISTRIB_FUNCTION_EXPONENTIAL: "AST_DISTRIB_FUNCTION_EXPONENTIAL",
    libsbml.AST_DISTRIB_FUNCTION_GAMMA: "AST_DISTRIB_FUNCTION_GAMMA",
    libsbml.AST_DISTRIB_FUNCTION_LAPLACE: "AST_DISTRIB_FUNCTION_LAPLACE",
    libsbml.AST_DISTRIB_FUNCTION_LOGNORMAL: "AST_DISTRIB_FUNCTION_LOGNORMAL",
    libsbml.AST_DISTRIB_FUNCTION_POISSON: "AST_DISTRIB_FUNCTION_POISSON",
    libsbml.AST_DISTRIB_FUNCTION_RAYLEIGH: "AST_DISTRIB_FUNCTION_RAYLEIGH",
    libsbml.AST_LINEAR_ALGEBRA_VECTOR: "AST_LINEAR_ALGEBRA_VECTOR",
    libsbml.AST_LINEAR_ALGEBRA_SELECTOR: "AST_LINEAR_ALGEBRA_SELECTOR",
    libsbml.AST_LINEAR_ALGEBRA_MATRIX: "AST_LINEAR_ALGEBRA_MATRIX",
    libsbml.AST_LINEAR_ALGEBRA_MATRIXROW: "AST_LINEAR_ALGEBRA_MATRIXROW",
    libsbml.AST_LINEAR_ALGEBRA_DETERMINANT: "AST_LINEAR_ALGEBRA_DETERMINANT",
    libsbml.AST_LINEAR_ALGEBRA_TRANSPOSE: "AST_LINEAR_ALGEBRA_TRANSPOSE",
    libsbml.AST_LINEAR_ALGEBRA_VECTOR_PRODUCT: "AST_LINEAR_ALGEBRA_VECTOR_PRODUCT",
    libsbml.AST_LINEAR_ALGEBRA_SCALAR_PRODUCT: "AST_LINEAR_ALGEBRA_SCALAR_PRODUCT",
    libsbml.AST_LINEAR_ALGEBRA_OUTER_PRODUCT: "AST_LINEAR_ALGEBRA_OUTER_PRODUCT",
    libsbml.AST_LOGICAL_EXISTS: "AST_LOGICAL_EXISTS",
    libsbml.AST_LOGICAL_FORALL: "AST_LOGICAL_FORALL",
    libsbml.AST_STATISTICS_MEAN: "AST_STATISTICS_MEAN",
    libsbml.AST_STATISTICS_MEDIAN: "AST_STATISTICS_MEDIAN",
    libsbml.AST_STATISTICS_MODE: "AST_STATISTICS_MODE",
    libsbml.AST_STATISTICS_MOMENT: "AST_STATISTICS_MOMENT",
    libsbml.AST_STATISTICS_SDEV: "AST_STATISTICS_SDEV",
    libsbml.AST_STATISTICS_VARIANCE: "AST_STATISTICS_VARIANCE",
    libsbml.AST_STATISTICS_MOMENTABOUT: "AST_STATISTICS_MOMENTABOUT",
    libsbml.AST_SERIES_PRODUCT: "AST_SERIES_PRODUCT",
    libsbml.AST_SERIES_SUM: "AST_SERIES_SUM",
    libsbml.AST_SERIES_LIMIT: "AST_SERIES_LIMIT",
    libsbml.AST_SERIES_TENDSTO: "AST_SERIES_TENDSTO",
    libsbml.AST_ALGEBRA_GCD: "AST_ALGEBRA_GCD",
    libsbml.AST_ALGEBRA_CONJUGATE: "AST_ALGEBRA_CONJUGATE",
    libsbml.AST_ALGEBRA_ARG: "AST_ALGEBRA_ARG",
    libsbml.AST_ALGEBRA_REAL: "AST_ALGEBRA_REAL",
    libsbml.AST_ALGEBRA_IMAGINARY: "AST_ALGEBRA_IMAGINARY",
    libsbml.AST_ALGEBRA_LCM: "AST_ALGEBRA_LCM",
    libsbml.AST_RELATIONS_EQUIVALENT: "AST_RELATIONS_EQUIVALENT",
    libsbml.AST_RELATIONS_APPROX: "AST_RELATIONS_APPROX",
    libsbml.AST_RELATIONS_FACTOROF: "AST_RELATIONS_FACTOROF",
    libsbml.AST_CALCULUS_INT: "AST_CALCULUS_INT",
    libsbml.AST_CALCULUS_DIFF: "AST_CALCULUS_DIFF",
    libsbml.AST_CALCULUS_PARTIALDIFF: "AST_CALCULUS_PARTIALDIFF",
    libsbml.AST_CALCULUS_LOWLIMIT: "AST_CALCULUS_LOWLIMIT",
    libsbml.AST_CALCULUS_UPLIMIT: "AST_CALCULUS_UPLIMIT",
    libsbml.AST_CALCULUS_DIVERGENCE: "AST_CALCULUS_DIVERGENCE",
    libsbml.AST_CALCULUS_GRAD: "AST_CALCULUS_GRAD",
    libsbml.AST_CALCULUS_CURL: "AST_CALCULUS_CURL",
    libsbml.AST_CALCULUS_LAPLACIAN: "AST_CALCULUS_LAPLACIAN",
    libsbml.AST_SET_THEORY_SET: "AST_SET_THEORY_SET",
    libsbml.AST_SET_THEORY_LIST: "AST_SET_THEORY_LIST",
    libsbml.AST_SET_THEORY_UNION: "AST_SET_THEORY_UNION",
    libsbml.AST_SET_THEORY_INTERSECT: "AST_SET_THEORY_INTERSECT",
    libsbml.AST_SET_THEORY_IN: "AST_SET_THEORY_IN",
    libsbml.AST_SET_THEORY_NOTIN: "AST_SET_THEORY_NOTIN",
    libsbml.AST_SET_THEORY_SUBSET: "AST_SET_THEORY_SUBSET",
    libsbml.AST_SET_THEORY_PRSUBSET: "AST_SET_THEORY_PRSUBSET",
    libsbml.AST_SET_THEORY_NOTSUBSET: "AST_SET_THEORY_NOTSUBSET",
    libsbml.AST_SET_THEORY_NOTPRSUBSET: "AST_SET_THEORY_NOTPRSUBSET",
    libsbml.AST_SET_THEORY_SETDIFF: "AST_SET_THEORY_SETDIFF",
    libsbml.AST_SET_THEORY_CARD: "AST_SET_THEORY_CARD",
    libsbml.AST_SET_THEORY_CARTESIANPRODUCT: "AST_SET_THEORY_CARTESIANPRODUCT",
    libsbml.AST_CONSTANT_IMAGINARYI: "AST_CONSTANT_IMAGINARYI",
    libsbml.AST_CONSTANTS_INTEGERS: "AST_CONSTANTS_INTEGERS",
    libsbml.AST_CONSTANTS_REALS: "AST_CONSTANTS_REALS",
    libsbml.AST_CONSTANTS_RATIONALS: "AST_CONSTANTS_RATIONALS",
    libsbml.AST_CONSTANTS_NATURALNUMBERS: "AST_CONSTANTS_NATURALNUMBERS",
    libsbml.AST_CONSTANTS_COMPLEXES: "AST_CONSTANTS_COMPLEXES",
    libsbml.AST_CONSTANTS_PRIMES: "AST_CONSTANTS_PRIMES",
    libsbml.AST_CONSTANTS_EMPTYSET: "AST_CONSTANTS_EMPTYSET",
    libsbml.AST_CONSTANTS_EULERGAMMA: "AST_CONSTANTS_EULERGAMMA",
    libsbml.AST_BASIC_CONTENT_INTERVAL: "AST_BASIC_CONTENT_INTERVAL",
    libsbml.AST_BASIC_CONTENT_INVERSE: "AST_BASIC_CONTENT_INVERSE",
    libsbml.AST_BASIC_CONTENT_CONDITION: "AST_BASIC_CONTENT_CONDITION",
    libsbml.AST_BASIC_CONTENT_DECLARE: "AST_BASIC_CONTENT_DECLARE",
    libsbml.AST_BASIC_CONTENT_COMPOSE: "AST_BASIC_CONTENT_COMPOSE",
    libsbml.AST_BASIC_CONTENT_IDENT: "AST_BASIC_CONTENT_IDENT",
    libsbml.AST_BASIC_CONTENT_DOMAIN: "AST_BASIC_CONTENT_DOMAIN",
    libsbml.AST_BASIC_CONTENT_CODOMAIN: "AST_BASIC_CONTENT_CODOMAIN",
    libsbml.AST_BASIC_CONTENT_IMAGE: "AST_BASIC_CONTENT_IMAGE",
    libsbml.AST_BASIC_CONTENT_DOMAINOFAPPLICATION: "AST_BASIC_CONTENT_DOMAINOFAPPLICATION",
    libsbml.AST_UNKNOWN: "AST_UNKNOWN",
}


class mathMLImporter:
    def __init__(self) -> None:
        self.mapper = ChainMap({}, mapper)

    def convert(self, node: libsbml.ASTNode):
        func = self.mapper[node.getType()]
        if isinstance(func, str):
            raise NotImplementedError(func)
        elif isinstance(func, dict):
            func = func[node.getName()]

        if node.getNumChildren() > 0:
            return func(*self.yield_children(node))
        elif callable(func):
            return func(node)
        else:
            return func

    def yield_children(self, node: libsbml.ASTNode) -> Real:
        for i in range(node.getNumChildren()):
            yield self.convert(node.getChild(i))

    def compile_function(self, func_name: str, node: libsbml.ASTNode):
        *params, body = self.yield_children(node)

        if any(keyword.iskeyword(p.name) for p in params):
            name_mapping = {p.name: p for p in params}
            for name in filter(keyword.iskeyword, name_mapping):
                new_name = name
                while new_name in name_mapping:
                    new_name = f"_{new_name}"
                name_mapping[name] = Real(new_name)
            params = name_mapping.values()
            body = body.subs_by_name(**name_mapping)

        func = as_function(
            body, func_name, tuple(map(str, params)), libsl=abstract
        )  # TODO: change to new symbolite
        self.add_function(func_name, func)
        return func

    def add_function(self, name: str, func):
        if self.mapper[libsbml.AST_FUNCTION] == "AST_FUNCTION":
            self.mapper[libsbml.AST_FUNCTION] = {}

        self.mapper[libsbml.AST_FUNCTION][name] = func



"""
simbio.reactions
~~~~~~~~~~~~~~~~

A reaction connects species to their rate of change.

:copyright: 2020 by SimBio Authors, see AUTHORS for more details.
:license: BSD, see LICENSE for more details.
"""

from __future__ import annotations

from .. import Parameter, Reactant, System, assign, reaction_initial
from .single import Conversion, Dissociation, Synthesis


class ReversibleSynthesis(System):
    """A Synthesis and Dissociation reactions.

    A + B <-> AB
    """

    A: Reactant = reaction_initial()
    B: Reactant = reaction_initial()
    AB: Reactant = reaction_initial()
    forward_rate: Parameter = assign()
    reverse_rate: Parameter = assign()

    forward_reaction = Synthesis(A=A, B=B, AB=AB, rate=forward_rate)
    backward_reaction = Dissociation(AB=AB, A=A, B=B, rate=reverse_rate)


class Equilibration(System):
    """A forward and backward Conversion reactions.

    A <-> B
    """

    A: Reactant = reaction_initial()
    B: Reactant = reaction_initial()
    forward_rate: Parameter = assign()
    reverse_rate: Parameter = assign()

    forward_reaction = Conversion(A=A, B=B, rate=forward_rate)
    backward_reaction = Conversion(A=B, B=A, rate=reverse_rate)


class CatalyzeConvert(System):
    """

    A + B <--> A:B --> P
    """

    A: Reactant = reaction_initial()
    B: Reactant = reaction_initial()
    AB: Reactant = reaction_initial()
    P: Reactant = reaction_initial()
    forward_rate: Parameter = assign()
    reverse_rate: Parameter = assign()
    conversion_rate: Parameter = assign()

    binding_reaction = ReversibleSynthesis(
        A=A, B=B, AB=AB, forward_rate=forward_rate, reverse_rate=reverse_rate
    )
    conversion_reaction = Conversion(A=AB, B=P, rate=conversion_rate)



from .. import (
    Parameter,
    RateLaw,
    Reactant,
    System,
    assign,
    reaction_initial,
)
from .compound import Dissociation, ReversibleSynthesis


class MichaelisMenten(System):
    E: Reactant = reaction_initial()
    S: Reactant = reaction_initial()
    ES: Reactant = reaction_initial()
    P: Reactant = reaction_initial()
    forward_rate: Parameter = assign()
    reverse_rate: Parameter = assign()
    catalytic_rate: Parameter = assign()

    binding_reaction = ReversibleSynthesis(
        A=E,
        B=S,
        AB=ES,
        forward_rate=forward_rate,
        reverse_rate=reverse_rate,
    )
    dissociation_reaction = Dissociation(
        AB=ES,
        A=E,
        B=P,
        rate=catalytic_rate,
    )

    def to_eq_approx(self):
        raise NotImplementedError
        # Should we generate a new S species or modify the concentration to add ES
        # Should we invalidate this reaction (self) because we are using these elsewhere
        # michaelis_constant = reverse_rate / forward_rate
        # return MichaelisMentenEqApprox(
        #     S=S,
        #     P=P,
        #     maximum_velocity=catalytic_rate * (E.concentration + ES.concentration),
        #     michaelis_constant=michaelis_constant,
        # )

    def to_qss_approx(self):
        raise NotImplementedError
        # Should we generate a new S species or modify the concentration to add ES
        # Should we invalidate this reaction (self) because we are using these elsewhere
        # michaelis_constant = (reverse_rate + catalytic_rate) / forward_rate
        # return MichaelisMentenEqApprox(
        #     S=S,
        #     P=P,
        #     maximum_velocity=catalytic_rate * (E.concentration + ES.concentration),
        #     michaelis_constant=michaelis_constant,
        # )


class MichaelisMentenEqApprox(System):
    S: Reactant = reaction_initial()
    P: Reactant = reaction_initial()
    maximum_velocity: Parameter = assign()
    dissociation_constant: Parameter = assign()
    reaction = RateLaw(
        reactants=[S],
        products=[P],
        rate_law=maximum_velocity * S.variable / (dissociation_constant + S.variable),
    )


class MichaelisMentenQuasiSSAprox(System):
    S: Reactant = reaction_initial()
    P: Reactant = reaction_initial()
    maximum_velocity: Parameter = assign()
    michaelis_constant: Parameter = assign()
    reaction = RateLaw(
        reactants=[S],
        products=[P],
        rate_law=maximum_velocity * S.variable / (michaelis_constant + S.variable),
    )



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto
from typing import NewType, Sequence

ID = NewType("ID", str)  # Identifier.
SBOTerm = NewType("SBOTerm", str)  # a 7-digit number, such as SBO:0000014


def is_valid_id(id: str) -> bool:
    """(alphanumeric | "_"), starting without a number."""
    raise NotImplementedError


@dataclass(kw_only=True)
class Math:
    pass


@dataclass(kw_only=True)
class Base:
    id: ID | None = None
    name: str | None = None
    meta_id: ID | None = None
    sbo_term: SBOTerm | None = None
    notes: str | None = None
    annotation: str | None = None


@dataclass(kw_only=True)
class Model(Base):
    substance_units: ID | None = None
    time_units: ID | None = None
    volume_units: ID | None = None
    area_units: ID | None = None
    length_units: ID | None = None
    extent_units: ID | None = None
    conversion_factor: ID | None = None
    functions: Sequence[FunctionDefinition] = ()
    units: Sequence[UnitDefinition] = ()
    compartments: Sequence[Compartment] = ()
    species: Sequence[Species] = ()
    parameters: Sequence[Parameter] = ()
    initial_assignments: Sequence[InitialAssignment] = ()
    rules: Sequence[Rule] = ()
    constraints: Sequence[Constraint] = ()
    reactions: Sequence[Reaction] = ()
    events: Sequence[Event] = ()


@dataclass(kw_only=True)
class FunctionDefinition(Base):
    id: ID
    math: Math


@dataclass(kw_only=True)
class Compartment(Base):
    id: ID
    spatial_dimensions: float | None = None
    size: float | None = None
    units: ID | None = None
    constant: bool


@dataclass(kw_only=True)
class Species(Base):
    id: ID
    compartment: ID
    initial_amount: float | None = None
    initial_concentration: float | None = None
    substance_units: ID | None = None
    has_only_substance_units: bool
    boundary_condition: bool
    constant: bool
    conversion_factor: ID | None = None


@dataclass(kw_only=True)
class Parameter(Base):
    id: ID
    value: float | None = None
    units: ID | None = None
    constant: bool


@dataclass(kw_only=True)
class LocalParameter(Base):
    id: ID
    value: float | None = None
    units: ID | None = None


@dataclass(kw_only=True)
class InitialAssignment(Base):
    "Mathematical expression used to determine the initial conditions of a model."

    symbol: ID
    math: Math


@dataclass(kw_only=True)
class Rule(Base):
    """A mathematical expression added to the set of equations."""

    math: Math


@dataclass(kw_only=True)
class AlgebraicRule(Rule):
    pass


@dataclass(kw_only=True)
class AssignmentRule(Rule):
    variable: ID


@dataclass(kw_only=True)
class RateRule(Rule):
    variable: ID


@dataclass(kw_only=True)
class Constraint(Base):
    """Defined by an arbitrary mathematical expression computing
    a True / False value from model symbols."""

    math: Math
    message: str


@dataclass(kw_only=True)
class Reaction(Base):
    id: ID
    reversible: bool
    fast: bool
    compartment: ID | None = None
    reactants: Sequence[SpeciesReference]
    products: Sequence[SpeciesReference]
    modifiers: list[ModifierSpeciesReference]
    kinetic_law: KineticLaw


@dataclass(kw_only=True)
class SimpleSpeciesReference(Base):
    species: ID


class ModifierSpeciesReference(SimpleSpeciesReference):
    pass


@dataclass(kw_only=True)
class SpeciesReference(SimpleSpeciesReference):
    stoichiometry: float | None = None
    constant: bool


@dataclass(kw_only=True)
class KineticLaw(Base):
    math: Math
    parameters: Sequence[Parameter]


@dataclass(kw_only=True)
class Event(Base):
    """Instantaneous, discontinuous change in one or more symbols of any type."""

    use_values_from_trigger_time: bool
    trigger: Trigger
    prority: Priority
    delay: Delay
    assignments: Sequence[EventAssignment]


@dataclass(kw_only=True)
class Trigger(Base):
    initial_value: bool
    persistent: bool
    math: Math


@dataclass(kw_only=True)
class Priority(Base):
    math: Math


@dataclass(kw_only=True)
class Delay(Base):
    math: Math


@dataclass(kw_only=True)
class EventAssignment(Base):
    math: Math


@dataclass(kw_only=True)
class UnitDefinition(Base):
    id: ID
    units: Sequence[Unit]

    def __post_init__(self):
        try:
            UnitKind[self.id]
        except KeyError:
            pass
        else:
            raise ValueError(f"{self.id} is the id of a base unit")


@dataclass(kw_only=True)
class Unit:
    """
    (multiplier * 10**scale * kind) ** exponent
    """

    kind: UnitKind
    exponent: float
    scale: int
    multiplier: float


class UnitKind(Enum):
    ampere = auto()
    coulomb = auto()
    gray = auto()
    joule = auto()
    litre = auto()
    mole = auto()
    radian = auto()
    steradian = auto()
    weber = auto()
    avogadro = auto()
    dimensionless = auto()
    henry = auto()
    katal = auto()
    lumen = auto()
    newton = auto()
    second = auto()
    tesla = auto()
    becquerel = auto()
    farad = auto()
    hertz = auto()
    kelvin = auto()
    lux = auto()
    ohm = auto()
    siemens = auto()
    volt = auto()
    candela = auto()
    gram = auto()
    item = auto()
    kilogram = auto()
    metre = auto()
    pascal = auto()
    sievert = auto()
    watt = auto()



import keyword
import math
from dataclasses import replace
from functools import reduce, singledispatchmethod
from operator import mul
from os import PathLike
from typing import Callable, Mapping, Sequence, TypeVar

import libsbml
import pint
from poincare.compile import depends_on_at_least_one_variable_or_time
from symbolite import Real, substitute

from ... import (
    Compartment,
    Constant,
    Independent,
    Parameter,
    RateLaw,
    Species,
    initial,
)
from ..mathML.importer import MathMLSpecialSymbol, MathMLSymbol
from . import from_libsbml, types

T = TypeVar("T")


class _Namespaces:
    SUPPORTED = {""}
    IGNORED = {"layout", "render"}
    UNSUPPORTED = {"comp", "fbc", "groups", "multi", "qual"}


def nan_to_none(x):
    if x is None or math.isnan(x):
        return None
    else:
        return x


class DynamicCompartment:
    def __init__(self, *, name_mapper: Callable[[str], str] = lambda x: x):
        self.name_mapping: dict[str, str] = {}
        self.name_mapper = name_mapper
        self.namespace = Compartment.__prepare__(None, ())
        self._annotations = self.namespace.setdefault("__annotations__", {})

    def build(self, name: str):
        return type(name, (Compartment,), self.namespace)

    def add(self, name: str, value, *, init: bool = True):
        self.name_mapping[name] = new_name = self.name_mapper(name)
        self.namespace[new_name] = value
        if init:
            self._annotations[new_name] = type(value)

    def __getattr__(self, name):
        try:
            new_name = self.name_mapping[name]
            return self.namespace[new_name]
        except KeyError as e:
            raise AttributeError(*e.args)


def load(
    file: PathLike,
    *,
    name: str | None = None,
    identity_mapper: Callable[[str], str] = lambda x: x,
):
    with open(file) as f:
        text = f.read()
    return loads(text, name=name, identity_mapper=identity_mapper)


def loads(
    sbml: str,
    *,
    name: str | None = None,
    identity_mapper: Callable[[str], str] = lambda x: x,
    ignore_namespaces: Sequence[str] = [],
):
    document: libsbml.SBMLDocument = libsbml.readSBMLFromString(sbml)
    if document.getNumErrors() != 0:
        raise RuntimeError("error reading the SBML file")

    ns: libsbml.XMLNamespaces = document.getSBMLNamespaces().getNamespaces()
    namespaces: set[str] = {ns.getPrefix(i) for i in range(ns.getLength())}
    namespaces.difference_update(_Namespaces.IGNORED)
    namespaces.difference_update(_Namespaces.SUPPORTED)
    namespaces.difference_update(ignore_namespaces)
    if len(namespaces) > 0:
        raise NotImplementedError(f"Unsupported SBML namespaces: {namespaces}")

    model: libsbml.Model = document.getModel()
    converted_model: types.Model = from_libsbml.Converter().convert(model)
    return convert(converted_model, name=name, identity_mapper=identity_mapper)


def convert(
    model: types.Model,
    *,
    name: str | None = None,
    identity_mapper: Callable[[str], str] = lambda x: x,
) -> type[Compartment]:
    if name is None:
        name = model.name
        if name is None:
            raise ValueError("must provide a name for the model")
    return SBMLImporter(model, identity_mapper=identity_mapper).simbio.build(name=name)


def _extra_check(func: Callable[[str], str]):
    def is_python_identifier(x: str) -> str:
        y = func(x)
        if keyword.iskeyword(y):
            return f".{y}"
        else:
            return y

    return is_python_identifier


class SBMLImporter:
    def __init__(
        self,
        model: types.Model,
        *,
        identity_mapper: Callable[[str], str] = lambda x: x,
        units: bool = False,
    ):
        if model.conversion_factor is not None:
            raise NotImplementedError("global conversion factor")

        self.model = model
        self.simbio = DynamicCompartment(name_mapper=_extra_check(identity_mapper))

        self.use_units = units
        if self.use_units:
            self.unit_registry = pint.get_application_registry()
            self.units = {
                u.id: reduce(mul, map(self.get_unit_from_registry, u.units))
                for u in model.units
            }

        self.species = {s.id: s for s in model.species}
        self.initials = {i.id: i for i in model.initial_assignments}
        self.algebraic_rules: Sequence[types.AlgebraicRule] = []
        self.assignment_rules: Mapping[types.ID, types.AssignmentRule] = {}
        self.rate_rules: Mapping[types.ID, types.RateRule] = {}
        for r in model.rules:
            match r:
                case types.AssignmentRule():
                    self.assignment_rules[r.variable] = r
                case types.RateRule():
                    self.rate_rules[r.variable] = r
                case types.AlgebraicRule():
                    self.algebraic_rules.append(r)
                case _:
                    raise NotImplementedError(type(r))

        for x in model.compartments:
            self.add_compartment(x)
        for x in model.parameters:
            self.add_parameter(x)
        for x in model.species:
            self.add_species(x)
        for x in model.initial_assignments:
            self.add_initial_assignment(x)
        for x in model.reactions:
            self.add_reaction(x)
        for r in model.rules:
            self.add(r)
        for a in model.constraints:
            self.add_constaint(a)
        for a in model.events:
            self.add_event(a)

    def get(self, item, default=None):
        match item:
            case Real(name=None):
                return item
            case MathMLSymbol(name=name):
                return getattr(self.simbio, name)
            case MathMLSpecialSymbol(name=name):
                return self.get_or_create_independent(name)
            case _:
                return item

    def get_unit_from_registry(self, u: types.Unit) -> pint.Quantity:
        unit = getattr(self.unit_registry, u.kind.name)
        return (u.multiplier * 10**u.scale * unit) ** u.exponent

    @singledispatchmethod
    def add(self, x):
        raise NotImplementedError(type(x))

    def get_or_create_independent(self, name: str):
        try:
            return getattr(self.simbio, name)
        except AttributeError:
            value = Independent(default=0)
            self.simbio.add(name, value)
            return value

    @add.register
    def add_compartment(self, c: types.Compartment):
        # if c.spatial_dimensions is not None:
        # raise NotImplementedError("spatial_dimensions in compartment")
        size = nan_to_none(c.size)
        if self.use_units and c.units is not None and size is not None:
            size *= self.units[c.units]

        if size != 1:
            raise NotImplementedError(f"compartment with size = {size} != 1.")

        self.simbio.add(c.id, Parameter(default=size))

    @add.register
    def add_parameter(self, p: types.Parameter):
        if p.id in self.initials or p.id in self.assignment_rules:
            default = None  # add assignment later
        else:
            default = nan_to_none(p.value)

        if self.use_units and p.units is not None and default is not None:
            default *= self.units[p.units]

        if p.id in self.rate_rules:
            value = initial(default=default)
        else:
            value = Parameter(default=default)

        self.simbio.add(p.id, value)

    @add.register
    def add_species(self, s: types.Species):
        # compartment: ID
        # has_only_substance_units: bool
        if s.conversion_factor is not None:
            raise NotImplementedError("conversion_factor in species")

        match [nan_to_none(s.initial_amount), nan_to_none(s.initial_concentration)]:
            case [None, None]:
                default = None
            case [default, None]:
                pass
            case [None, default]:
                if self.use_units and s.substance_units is not None:
                    default *= self.units[s.substance_units]
            case _:
                raise ValueError(
                    f"both amount an concentration specified for Species {s.id}"
                )

        if s.id in self.assignment_rules:
            value = Parameter(default=default)
        else:
            value = initial(default=default)
        self.simbio.add(s.id, value)

    def get_symbol(self, name: str, expected_type: type[T] = object) -> T:
        value = getattr(self.simbio, name)
        if not isinstance(value, expected_type):
            raise TypeError(f"unexpected type: {type(value)}")
        return value

    def get_species_reference(self, s: types.SimpleSpeciesReference) -> Species:
        # s.constant: bool
        species = self.get_symbol(s.species, Species)
        if isinstance(s, types.SpeciesReference) and s.stoichiometry is not None:
            return Species(species, s.stoichiometry)
        else:
            return Species(species)

    @add.register
    def add_reaction(self, r: types.Reaction):
        if r.fast:
            raise NotImplementedError("fast reactions")

        # r.compartment
        def yield_species(references: Sequence[types.SimpleSpeciesReference]):
            for r in references:
                s = self.species[r.species]
                if not (
                    s.boundary_condition or s.constant or s.id in self.assignment_rules
                ):
                    yield self.get_species_reference(r)

        reactants = list(yield_species(r.reactants))
        products = list(yield_species(r.products))
        modifiers = list(yield_species(r.modifiers))
        for m in modifiers:
            reactants.append(m)
            products.append(m)
        kinetic_law = r.kinetic_law
        formula: Real = kinetic_law.math
        if len(kinetic_law.parameters) > 0:
            mapping = {}
            for p in kinetic_law.parameters:
                new_id = f"{r.id}__{p.id}"
                self.add_parameter(replace(p, id=new_id))
                mapping[p.id] = MathMLSymbol(new_id)  # TODO: replace substitute by name
            # formula = substitute_by_name(
            #     formula, **mapping
            # )
        # formula = substitute(formula, GetAsVariable(self.get))

        self.simbio.add(
            r.id,
            RateLaw(reactants=reactants, products=products, rate_law=formula),
        )
        return

        # Reversible equations
        # if not r.reversible:
        #     self.simbio.add(
        #         r.id,
        #         RateLaw(reactants=reactants, products=products, rate_law=formula),
        #     )
        #     return

        # match formula:
        #     case Symbol(
        #         expression=symbol.Expression(func=symbol.sub, args=(forward, reverse))
        #     ):
        #         pass
        #     case Symbol(
        #         expression=symbol.Expression(
        #             func=symbol.mul,
        #             args=(
        #                 compartment,
        #                 Symbol(
        #                     expression=symbol.Expression(
        #                         func=symbol.sub, args=(forward, reverse)
        #                     )
        #                 ),
        #             )
        #             | (
        #                 Symbol(
        #                     expression=symbol.Expression(
        #                         func=symbol.sub, args=(forward, reverse)
        #                     )
        #                 ),
        #                 compartment,
        #             ),
        #         )
        #     ):
        #         forward = compartment * forward
        #         reverse = compartment * reverse
        #     case _:
        #         # Cannot split formula into forward and reverse
        #         self.simbio.add(
        #             r.id,
        #             RateLaw(reactants=reactants, products=products, rate_law=formula),
        #         )
        #         return

    @add.register
    def add_initial_assignment(self, a: types.InitialAssignment):
        component = self.get_symbol(a.symbol)
        value = substitute(a.math, GetAsVariable(self.get))

        if depends_on_at_least_one_variable_or_time(value):
            raise NotImplementedError("initial conditions must by constant")

        if value is None:
            return
        if isinstance(component, Species):
            component.initial = value
        elif isinstance(component, Constant | Parameter):
            component.default = value
        else:
            raise TypeError(
                f"cannot perform initial assignment for {a.symbol} of type {type(component)}"
            )

    @add.register
    def add_assignment_rule(self, r: types.AssignmentRule):
        component = self.get_symbol(r.variable, Parameter)
        if any(r.variable == c.id for c in self.model.compartments):
            raise NotImplementedError("assignment rules for compartments")

        value = substitute(r.math, GetAsVariable(self.get))
        if value is None:
            return
        elif isinstance(value, RateLaw):
            raise NotImplementedError("RateRule in AssignmentRule")
        else:
            component.default = value

    @add.register
    def add_rate_rule(self, r: types.RateRule):
        species = self.get_symbol(r.variable, Species)
        value: Real = substitute(r.math, GetAsVariable(self.get))
        if value is None:
            return
        elif isinstance(value, RateLaw):
            raise NotImplementedError("RateRule in AssignmentRule")
        else:
            eq = species.derive() << value

        name = r.id
        if name is None or name == r.variable:
            name = f"{r.variable}__rate_rule"
        self.simbio.add(name, eq)

    @add.register
    def add_algebraic_rule(self, r: types.AlgebraicRule):
        raise NotImplementedError("algebraic rules are not yet supported")

    @add.register
    def add_event(self, r: types.Event):
        raise NotImplementedError("Event")

    @add.register
    def add_constaint(self, r: types.Constraint):
        raise NotImplementedError("Constraint")


class GetAsVariable:
    def __init__(self, getter):
        self.getter = getter

    def get(self, key, default=None):
        value = self.getter(key, default=None)
        return value



from __future__ import annotations

from functools import partial, reduce
from typing import Iterator, Protocol

from symbolite.abstract import real

from .symbol import MathMLSpecialSymbol, MathMLSymbol


class Element(Protocol):
    @property
    def tag(self) -> str: ...

    @property
    def attrib(self) -> dict[str, str]: ...

    @property
    def text(self) -> str: ...

    def __iter__(self) -> Iterator[Element]: ...


def star_reduce(func, *args):
    return reduce(func, args)


tags = {
    "cn": MathMLSymbol,
    "ci": MathMLSpecialSymbol,
    "csymbol": MathMLSpecialSymbol,
    "sep": NotImplemented,
    "apply": NotImplemented,
    "piecewise": NotImplemented,
    "piece": NotImplemented,
    "otherwise": NotImplemented,
    "lambda": NotImplemented,
    "eq": real.eq,
    "neq": real.ne,
    "gt": real.gt,
    "lt": real.lt,
    "geq": real.ge,
    "leq": real.le,
    "plus": partial(star_reduce, real.add),
    "minus": partial(star_reduce, real.sub),
    "times": partial(star_reduce, real.mul),
    "divide": partial(star_reduce, real.truediv),
    "power": real.pow,
    "root": NotImplemented,
    "abs": real.abs,
    "exp": real.exp,
    "ln": real.log,
    "log": real.log10,
    "floor": real.floor,
    "ceiling": real.ceil,
    "factorial": real.factorial,
    "quotient": NotImplemented,
    "max": NotImplemented,
    "min": NotImplemented,
    "rem": real.mod,
    "and": partial(star_reduce, real.and_),
    "or": partial(star_reduce, real.or_),
    "xor": real.xor,
    "not": real.invert,
    "implies": NotImplemented,
    "degree": NotImplemented,
    "bvar": NotImplemented,
    "logbase": NotImplemented,
    "sin": real.sin,
    "cos": real.cos,
    "tan": real.tan,
    "sec": NotImplemented,
    "csc": NotImplemented,
    "cot": NotImplemented,
    "sinh": real.sinh,
    "cosh": real.cosh,
    "tanh": real.tanh,
    "sech": NotImplemented,
    "csch": NotImplemented,
    "coth": NotImplemented,
    "arcsin": real.asin,
    "arccos": real.acos,
    "arctan": real.atan,
    "arcsec": NotImplemented,
    "arccsc": NotImplemented,
    "arccot": NotImplemented,
    "arcsinh": real.asinh,
    "arccosh": real.acosh,
    "arctanh": real.atanh,
    "arcsech": NotImplemented,
    "arccsch": NotImplemented,
    "arccoth": NotImplemented,
    "true": True,
    "false": False,
    "notanumber": real.nan,
    "pi": real.pi,
    "infinity": real.inf,
    "exponentiale": real.e,
    "semantics": NotImplemented,
    "annotation": NotImplemented,
    "annotation-xml": NotImplemented,
    "math": lambda x: x,
}

ns_map = {"http://www.w3.org/1998/Math/MathML": tags}


def from_element(element: Element) -> Real:
    result = parse(element)
    if isinstance(result, list) and len(result) == 1:
        return result[0]
    raise RuntimeError("unexpected result when parsing mathML")


def _namespace_and_tag(x: str) -> tuple[str, str]:
    if x.startswith("{"):
        ns, _, tag = x[1:].partition("}")
        return ns, tag
    else:
        return "", x


def parse(element: Element):
    ns, tag = _namespace_and_tag(element.tag)
    cls = ns_map[ns][tag]

    if tag == "apply":
        func, *children = (parse(child) for child in element)
        return func(*children)

    children = [parse(child) for child in element]
    if len(children) > 0:
        return children
    elif (text := element.text) is not None:
        return cls(text.strip())
    else:
        return cls



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
)

from poincare import (
    Derivative,
    Variable,
)
from poincare._node import Node, NodeMapper
from poincare.reactions import (
    Reactant,
)
from poincare.reactions.reactions import (
    compensate_volume,
    make_concentration,
)
from poincare.types import (
    Initial,
    System,
)
from symbolite import Real, substitute

if TYPE_CHECKING:
    pass


def is_instance_or_subclass(obj: Any, cls: type):
    try:
        return issubclass(obj, cls) or isinstance(obj, cls)
    except TypeError:
        return isinstance(obj, cls)


def first_system_parent(obj: Node | None) -> System | type[System] | None:
    if isinstance(obj, Node):
        parent = getattr(obj, "parent", None)
        if is_instance_or_subclass(parent, System):
            return parent
        else:
            return first_system_parent(parent)
    else:
        return None


def concentration(*, default: Initial | None = None):
    return Species(initial=default, concentration=True)


def amount(*, default: Initial | None = None):
    return Species(initial=default, concentration=False)


def reaction_concentration(*, default: Initial | None = None):
    return Reactant(
        variable=Species(initial=default, concentration=True), stoichiometry=1
    )


def reaction_amount(*, default: Initial | None = None):
    return Reactant(
        variable=Species(initial=default, concentration=False), stoichiometry=1
    )


def volume(*, default: Initial | None = None):
    return Volume(initial=default)


class Species(Variable):
    def __init__(self, *, initial: Initial, concentration: bool):
        self.concentration = concentration
        super().__init__(initial=initial)

    def _copy_from(self, parent: Node):
        mapper = NodeMapper(parent)
        copy = self.__class__(
            initial=substitute(self.initial, mapper), concentration=self.concentration
        )
        for order, der in self.derivatives.items():
            copy.derivatives[order] = Derivative(
                copy, initial=substitute(der.initial, mapper), order=order
            )
        copy.equation_order = self.equation_order
        return copy


@compensate_volume.register
def compensate_volume_Species(
    species: Species, rhs: Real | Initial, reaction_is_concentration: bool
) -> Real | Initial:
    system_parent = first_system_parent(species)
    if is_instance_or_subclass(system_parent, Compartment):
        if species.concentration and not reaction_is_concentration:
            return rhs / system_parent._simbio_volume
        elif not species.concentration and reaction_is_concentration:
            return rhs * system_parent._simbio_volume
        else:
            return rhs
    else:
        return rhs


@make_concentration.register
def make_concentration_Species(species: Species) -> Real | Initial:
    system_parent = first_system_parent(species)
    if not species.concentration and (
        is_instance_or_subclass(system_parent, Compartment)
    ):
        try:
            return species / system_parent._simbio_volume
        except AttributeError as err:
            if not hasattr(system_parent, "_simbio_volume"):
                raise AttributeError("Compartments must have a Volume")
            else:
                raise err
    else:
        return species


# def verify_species_units(initial):
#     if isinstance(initial, pint.Quantity):
#         if (
#             initial.dimensionality == concetration_dim
#             or initial.dimensionality == substance_dim
#         ):
#             return
#         else:
#             raise TypeError(
#                 f"Species must have concentration or substance dimensionality"
#             )
#     else:
#         raise TypeError(
#             f"Species does not have units, must have units of substance or concentration"
#         )


class Volume(Variable):
    def __set_name__(self, cls: Node, name: str):
        super().__set_name__(cls, name)
        if cls is not None:
            setattr(cls, "_simbio_volume", self)


class Compartment(System, abstract=True):
    def __init_subclass__(cls, **kwargs) -> None:
        volumes = set(cls._yield(Volume, recursive=False))
        if len(volumes) == 0:
            raise AttributeError("Compartments must have a Volume")
        elif len(volumes) > 1:
            raise AttributeError("Compartments can only have one Volume")
        super().__init_subclass__(**kwargs)

    def __init__(self, *args, **kwargs):
        for key, value in kwargs.items():
            if isinstance(getattr(self.__class__, key, None), Species | Volume):
                if isinstance(value, Node):
                    raise TypeError(
                        "Compartment Species and Volume cannot be linked to external variables, only initials can be passed on instantiation"
                    )
        super().__init__(*args, **kwargs)



from poincare import (
    Constant,
    Independent,
    Parameter,
    Simulator,
    Variable,
    assign,
    initial,
)
from poincare.reactions import (
    AbsoluteRateLaw,
    MassAction,
    RateLaw,
    Reactant,
    reaction_initial,
)

from .core import (
    Compartment,
    Species,
    System,
    Volume,
    amount,
    concentration,
    reaction_amount,
    reaction_concentration,
    volume,
)

__all__ = [
    "Constant",
    "Independent",
    "assign",
]
__all__ += [
    "Simulator",
    "Compartment",
    "System",
    "Variable",
    "Reactant",
    "Species",
    "Volume",
    "Parameter",
    "MassAction",
    "RateLaw",
    "AbsoluteRateLaw",
    "initial",
    "amount",
    "concentration",
    "reaction_initial",
    "reaction_concentration",
    "reaction_amount",
    "volume",
]



from __future__ import annotations

from .. import (
    MassAction,
    Parameter,
    Reactant,
    System,
    assign,
    reaction_initial,
)


class Creation(System):
    """A substance is created from nothing at a constant rate.

    ∅ -> A
    """

    A: Reactant = reaction_initial()
    rate: Parameter = assign()
    reaction = MassAction(reactants=[], products=[A], rate=rate)


class AutoCreation(System):
    """A substance is created at a rate proportional to its abundance.

    A -> 2A
    """

    A: Reactant = reaction_initial()
    rate: Parameter = assign()
    reaction = MassAction(reactants=[A], products=[2 * A], rate=rate)


class Destruction(System):
    """A substance degrades into nothing.

    A -> ∅
    """

    A: Reactant = reaction_initial()
    rate: Parameter = assign()
    reaction = MassAction(reactants=[A], products=[], rate=rate)


class Conversion(System):
    """A substance convert to another.

    A -> B
    """

    A: Reactant = reaction_initial()
    B: Reactant = reaction_initial()
    rate: Parameter = assign()
    reaction = MassAction(reactants=[A], products=[B], rate=rate)


class Synthesis(System):
    """Two or more simple substances combine to form a more complex substance.

    A + B -> AB
    """

    A: Reactant = reaction_initial()
    B: Reactant = reaction_initial()
    AB: Reactant = reaction_initial()
    rate: Parameter = assign()
    reaction = MassAction(reactants=[A, B], products=[AB], rate=rate)


class Dissociation(System):
    """A more complex substance breaks down into its more simple parts.

    AB -> A + B
    """

    AB: Reactant = reaction_initial()
    A: Reactant = reaction_initial()
    B: Reactant = reaction_initial()
    rate: Parameter = assign()
    reaction = MassAction(reactants=[AB], products=[A, B], rate=rate)



import numpy as np
from poincare import Independent, System, Variable

from . import (
    AbsoluteRateLaw,
    Compartment,
    MassAction,
    Parameter,
    RateLaw,
    Simulator,
    Species,
    assign,
    Reactant,
    reaction_initial,
)
from .core import (
    Volume,
    amount,
    compensate_volume,
    concentration,
    make_concentration,
    reaction_amount,
    reaction_concentration,
    volume,
)


def test_no_external_species_in_nested_compartment():
    class Nested(Compartment):
        V: Volume = volume(default=1)
        A: Species = amount(default=1)

        eq = RateLaw(reactants=[A], products=[2 * A], rate_law=1)

    try:

        class Model(Compartment):
            V: Volume = volume(default=1)
            A: Species = amount(default=1)

            nested = Nested(A=A)
            eq = RateLaw(reactants=[A], products=[2 * A], rate_law=1)
    except TypeError:
        assert True
        return
    assert False


def test_external_ic_in_nested_compartment():
    class Nested(Compartment):
        V: Volume = volume(default=1)
        A: Species = amount(default=1)

        eq = RateLaw(reactants=[A], products=[2 * A], rate_law=1)

    class Model(Compartment):
        V: Volume = volume(default=1)
        A: Species = amount(default=1)

        nested = Nested(A=2)
        eq = RateLaw(reactants=[A], products=[2 * A], rate_law=1)

    assert Model.nested.A.initial == 2


def test_no_volume_in_compartment():
    try:

        class Model(Compartment):
            A: Species = amount(default=1)

            eq = RateLaw(reactants=[A], products=[2 * A], rate_law=1)
    except AttributeError:
        assert True
        return
    assert False


def test_more_than_one_volume_in_compartment():
    try:

        class Model(Compartment):
            V1: Volume = volume(default=1)
            V2: Volume = volume(default=1)
            A: Species = amount(default=1)

            eq = RateLaw(reactants=[A], products=[2 * A], rate_law=1)
    except AttributeError:
        assert True
        return
    assert False


def test_rate_law_with_amount():
    class Model(System):
        A: Variable = Variable(initial=1)
        B: Variable = Variable(initial=2)
        AB: Variable = Variable(initial=0)

        eq = RateLaw(reactants=[A, 2 * B], products=[AB], rate_law=2)

    sim1 = Simulator(Model)
    result_1 = sim1.solve(save_at=np.linspace(0, 10, 10))

    class VolumeModel(Compartment):
        V: Volume = volume(default=2)
        A: Species = amount(default=1)
        B: Species = amount(default=2)
        AB: Species = amount(default=0)

        eq = RateLaw(reactants=[A, 2 * B], products=[AB], rate_law=1)

    sim2 = Simulator(VolumeModel)
    result_2 = sim2.solve(save_at=np.linspace(0, 10, 10))[["A", "AB", "B"]]
    assert np.all((result_1 == result_2).to_array().to_numpy())


def test_rate_law_with_concentration():
    class Model(System):
        A: Variable = Variable(initial=1)
        B: Variable = Variable(initial=2)
        AB: Variable = Variable(initial=0)

        eq = RateLaw(reactants=[A, 2 * B], products=[AB], rate_law=1)

    sim1 = Simulator(Model)
    result_1 = sim1.solve(save_at=np.linspace(0, 10, 10))

    class VolumeModel(Compartment):
        V: Volume = volume(default=2)
        A: Species = concentration(default=1)
        B: Species = concentration(default=2)
        AB: Species = concentration(default=0)

        eq = RateLaw(reactants=[A, 2 * B], products=[AB], rate_law=1)

    sim2 = Simulator(VolumeModel)
    result_2 = sim2.solve(save_at=np.linspace(0, 10, 10))[["A", "AB", "B"]]
    assert np.all((result_1 == result_2).to_array().to_numpy())


def test_absolute_rate_law_with_amount():
    class Model(System):
        A: Variable = Variable(initial=1)
        B: Variable = Variable(initial=2)
        AB: Variable = Variable(initial=0)

        eq = AbsoluteRateLaw(reactants=[A, 2 * B], products=[AB], rate_law=1)

    sim1 = Simulator(Model)
    result_1 = sim1.solve(save_at=np.linspace(0, 10, 10))

    class VolumeModel(Compartment):
        V: Volume = volume(default=2)
        A: Species = amount(default=1)
        B: Species = amount(default=2)
        AB: Species = amount(default=0)

        eq = AbsoluteRateLaw(reactants=[A, 2 * B], products=[AB], rate_law=1)

    sim2 = Simulator(VolumeModel)
    result_2 = sim2.solve(save_at=np.linspace(0, 10, 10))[["A", "AB", "B"]]
    assert np.all((result_1 == result_2).to_array().to_numpy())


def test_absolute_rate_law_with_concentration():
    class Model(System):
        A: Variable = Variable(initial=1)
        B: Variable = Variable(initial=2)
        AB: Variable = Variable(initial=0)

        eq = AbsoluteRateLaw(reactants=[A, 2 * B], products=[AB], rate_law=1 / 2)

    sim1 = Simulator(Model)
    result_1 = sim1.solve(save_at=np.linspace(0, 10, 10))

    class VolumeModel(Compartment):
        V: Volume = volume(default=2)
        A: Species = concentration(default=1)
        B: Species = concentration(default=2)
        AB: Species = concentration(default=0)

        eq = AbsoluteRateLaw(reactants=[A, 2 * B], products=[AB], rate_law=1)

    sim2 = Simulator(VolumeModel)
    result_2 = sim2.solve(save_at=np.linspace(0, 10, 10))[["A", "AB", "B"]]
    assert np.all((result_1 == result_2).to_array().to_numpy())


def test_mass_action_with_amount():
    class Model(System):
        A: Variable = Variable(initial=1)
        B: Variable = Variable(initial=2)
        AB: Variable = Variable(initial=0)

        eq = RateLaw(
            reactants=[A, 2 * B], products=[AB], rate_law=(A / 2) * (B / 2) ** 2 * 2
        )

    sim1 = Simulator(Model)
    result_1 = sim1.solve(save_at=np.linspace(0, 10, 10))

    class VolumeModel(Compartment):
        V: Volume = volume(default=2)
        A: Species = amount(default=1)
        B: Species = amount(default=2)
        AB: Species = amount(default=0)

        eq = MassAction(reactants=[A, 2 * B], products=[AB], rate=1)

    sim2 = Simulator(VolumeModel)
    result_2 = sim2.solve(save_at=np.linspace(0, 10, 10))[["A", "AB", "B"]]
    assert np.all((result_1 == result_2).to_array().to_numpy())


def test_mass_action_with_concentration():
    class Model(System):
        A: Variable = Variable(initial=1)
        B: Variable = Variable(initial=2)
        AB: Variable = Variable(initial=0)

        eq = RateLaw(reactants=[A, 2 * B], products=[AB], rate_law=A * B**2)

    sim1 = Simulator(Model)
    result_1 = sim1.solve(save_at=np.linspace(0, 10, 10))

    class VolumeModel(Compartment):
        V: Volume = volume(default=2)
        A: Species = concentration(default=1)
        B: Species = concentration(default=2)
        AB: Species = concentration(default=0)

        eq = MassAction(reactants=[A, 2 * B], products=[AB], rate=1)

    sim2 = Simulator(VolumeModel)
    result_2 = sim2.solve(save_at=np.linspace(0, 10, 10))[["A", "AB", "B"]]
    assert np.all((result_1 == result_2).to_array().to_numpy())


def test_changing_volume():
    class Model(System):
        t: Independent = Independent()
        V: Parameter = assign(default=t**2 / 2 + 1)
        A: Variable = Variable(initial=1)
        B: Variable = Variable(initial=2)
        AB: Variable = Variable(initial=0)

        eq = RateLaw(reactants=[A, 2 * B], products=[AB], rate_law=A * B**2 / V)

    sim1 = Simulator(Model)
    result_1 = sim1.solve(save_at=np.linspace(0, 10, 10))

    class VolumeModel(Compartment):
        t: Independent = Independent()
        V: Volume = volume(default=1)
        A: Species = concentration(default=1)
        B: Species = concentration(default=2)
        AB: Species = concentration(default=0)

        eq = MassAction(reactants=[A, 2 * B], products=[AB], rate=1)
        vol_eq = V.derive() << t

    sim2 = Simulator(VolumeModel)
    result_2 = sim2.solve(save_at=np.linspace(0, 10, 10))
    calculated_volumes = np.linspace(0, 10, 10) ** 2 / 2 + 1
    volumes = np.asarray(result_2["V"])
    assert np.all(
        np.abs((result_1 - result_2[["A", "AB", "B"]]))
        <= (result_1 + result_2[["A", "AB", "B"]]) / 2 * 0.01
    )
    assert np.all(
        np.abs(calculated_volumes - volumes)
        <= (calculated_volumes + volumes) / 2 * 0.01
    )


def test_nested_compartments():
    class Nested(Compartment):
        V: Volume = volume(default=1)
        A: Species = amount(default=1)

        eq = RateLaw(reactants=[A], products=[2 * A], rate_law=1)

    class Model(Compartment):
        V: Volume = volume(default=1)
        A: Species = amount(default=1)

        nested = Nested()
        eq = RateLaw(reactants=[A], products=[2 * A], rate_law=1)

    sim = Simulator(Model)
    sim.solve(save_at=np.linspace(0, 10, 10))

    assert Model.A and Model.nested.A in sim.compiled.variables
    assert Model._simbio_volume == Model.V
    assert Model.nested._simbio_volume == Model.nested.V


def test_species_in_reactant():
    class Nested(Compartment):
        V: Volume = Volume(initial=2)
        A: Reactant = reaction_amount(default=0.5)
        B: Reactant = reaction_concentration(default=2)
        AB: Reactant = reaction_concentration(default=0)

        eq = RateLaw(reactants=[A, B], products=[AB], rate_law=2)

    assert make_concentration(Nested.A.variable) == Nested.A.variable / Nested.V
    assert make_concentration(Nested.B.variable) == Nested.B.variable
    assert (
        compensate_volume(
            Nested.A.variable, rhs=2 * Nested.A.variable, reaction_is_concentration=True
        )
        == 2 * Nested.A.variable * Nested.V
    )
    assert (
        compensate_volume(
            Nested.B.variable, rhs=2 * Nested.B.variable, reaction_is_concentration=True
        )
        == 2 * Nested.B.variable
    )
    nsim = Simulator(Nested)
    nsim.solve(save_at=np.linspace(0, 10, 10))

    class Model(Compartment):
        V: Volume = Volume(initial=4)
        A: Reactant = reaction_amount(default=1)
        B: Reactant = reaction_concentration(default=3)
        nested = Nested()
        eq = MassAction(products=[A, B], reactants=[nested.AB], rate=1)

    sim = Simulator(Model)
    sim.solve(save_at=np.linspace(0, 10, 10))


def test_species_in_reactant_with_external_stoichiometry():
    class Nested(System):
        A: Reactant = reaction_initial(default=0.5)
        B: Reactant = reaction_initial(default=2)
        AB: Reactant = reaction_initial(default=0)

        eq = RateLaw(reactants=[A, B], products=[AB], rate_law=2)

    nsim = Simulator(Nested)
    nsim.solve(save_at=np.linspace(0, 10, 10))

    class Model(Compartment):
        V: Volume = Volume(initial=4)
        A: Reactant = reaction_amount(default=1)
        B: Reactant = reaction_concentration(default=3)
        nested = Nested(A=2 * A, B=3 * B)

        eq = MassAction(products=[nested.A, B], reactants=[nested.AB], rate=1)

    assert Model.nested.A.stoichiometry == 2
    assert Model.nested.B.stoichiometry == 3
    assert (
        make_concentration(Model.nested.A.variable) == Model.nested.A.variable / Model.V
    )
    assert make_concentration(Model.nested.B.variable) == Model.nested.B.variable
    assert (
        compensate_volume(
            Model.nested.A.variable,
            2 * Model.nested.A.variable,
            reaction_is_concentration=True,
        )
        == 2 * Model.nested.A.variable * Model.V
    )
    assert (
        compensate_volume(
            Model.nested.B.variable,
            2 * Model.nested.B.variable,
            reaction_is_concentration=True,
        )
        == 2 * Model.nested.B.variable
    )
    sim = Simulator(Model)
    sim.solve(save_at=np.linspace(0, 10, 10))



from pytest import mark
from symbolite.abstract import real

from . import mathMLImporter, to_mathML
from .symbol import MathMLSymbol as Symbol

x, y = map(Symbol, ["x", "y"])


@mark.parametrize(
    "expr",
    [
        1,
        x,
        x * y,
        x + y,
        x * 2,
        2 * x,
        2 * x + y,
        x**2,
        x**0.5,
        real.cos(x),
        real.sqrt(x),
        x < 1,
        x < y,
        ~x,
    ],
)
def test_mathML_roundtrip(expr: Symbol):
    node = to_mathML(expr)
    expr2 = mathMLImporter().convert(node)
    assert expr2 == expr



from typing import assert_never

from networkx import DiGraph, Graph, bipartite

from ...core import Compartment, RateLaw, Species


def graph(model: type[Compartment], /) -> DiGraph:
    """Construct a directed bipartite graph of Species and Reactions."""
    g = DiGraph()
    for x in model._yield(Species | RateLaw):  # type: ignore
        match x:
            case Species():
                g.add_node(x)
            case RateLaw():
                name = str(x)
                g.add_node(name, reaction=x)
                for r in x.reactants:
                    g.add_edge(r.variable, name, stoichiometry=r.stoichiometry)
                for p in x.products:
                    g.add_edge(name, p.variable, stoichiometry=p.stoichiometry)
            case _:
                assert_never(x)
    return g


def to_species_graph(graph: DiGraph, /) -> Graph:
    """Project a bipartite graph into the Species graph.

    Two Species are connected if they take part in the same Reaction.
    """
    return bipartite.projected_graph(
        graph,
        nodes=[n for n in graph if isinstance(n, Species)],
    )


def to_reaction_graph(graph: DiGraph, /) -> Graph:
    """Project a bipartite graph into the Reaction graph.

    Two Reactions are connected if they share a common Species.
    """
    return bipartite.projected_graph(
        graph,
        nodes=[n for n in graph if not isinstance(n, Species)],
    )



import zipfile
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable

import numpy as np
import pandas as pd
import pooch
from pandas.testing import assert_frame_equal
from poincare.solvers import LSODA
from pydantic import BaseModel
from pytest import mark, xfail
from symbolite.core import Unsupported

from .... import Simulator
from .. import loads

PATH = pooch.retrieve(
    "https://github.com/sbmlteam/sbml-test-suite/releases/download/3.4.0/semantic_tests.v3.4.0.zip",
    known_hash="sha256:ec1abd53e28221b8828c88f237142067d0e6e63042519f486868195b581c4b2c",
)
FILE = zipfile.Path(PATH, "semantic/")
MODELS = sorted(p.name for p in FILE.iterdir() if p.is_dir())
ATOL_FACTOR: dict[str, float] = pd.read_csv(
    Path(__file__).parent / "atol_factor.csv",
    header=None,
    index_col=0,
    dtype={0: str, 1: float},
).squeeze()


class Settings(BaseModel):
    start: float | None
    duration: float | None
    steps: int | None
    variables: list[str]
    absolute: float
    relative: float
    amount: list[str]
    concentration: list[str]

    @classmethod
    def from_file(cls, file: Iterable[str], /):
        lists = ("variables", "amount", "concentration")

        kwargs = {}
        for line in file:
            k, _, v = line.partition(":")
            if k in lists:
                v = [v.strip() for v in v.split(",")]
            else:
                v = v.strip()
                if v == "":
                    v = None
            kwargs[k] = v
        return cls(**kwargs)


@dataclass
class Loader:
    file: zipfile.Path
    model: str

    def read_sbml(self, lv: tuple[int, int] | None = None) -> str:
        if lv is not None:
            level, version = lv
            relative_path = f"{self.model}/{self.model}-sbml-l{level}v{version}.xml"
            return (self.file / relative_path).read_text()

        for lv in [(3, 2), (3, 1), (2, 5), (2, 4), (2, 3), (2, 2), (2, 1)]:
            try:
                return self.read_sbml(lv)
            except FileNotFoundError as e:
                last_error = e
        else:
            raise last_error

    def read_settings(self) -> Settings:
        relative_path = f"{self.model}/{self.model}-settings.txt"
        with (self.file / relative_path).open() as f:
            return Settings.from_file(f)

    def read_results(self) -> pd.DataFrame:
        relative_path = f"{self.model}/{self.model}-results.csv"
        with (self.file / relative_path).open() as f:
            return pd.read_csv(f)


@mark.parametrize("model", MODELS)
def test_loading(model):
    sbml = Loader(FILE, model).read_sbml()
    try:
        model = loads(sbml, name=model)
    except Exception as e:
        xfail(str(e))


@mark.parametrize("model_id", MODELS)
def test_running(model_id):
    loader = Loader(FILE, model_id)

    sbml = loader.read_sbml()
    try:
        model = loads(sbml, name=model_id)
    except Exception:
        # There is a separate test for model loading errors
        xfail("Loading error")

    if len(model.variables) == 0:
        xfail("no variables in model")

    settings = loader.read_settings()
    if settings.start is None:
        xfail("start is None")
    elif settings.duration is None:
        xfail("duration is None")
    elif settings.steps is None:
        xfail("steps is None")
    else:
        save_at = np.linspace(
            settings.start,
            settings.start + settings.duration,
            settings.steps + 1,
        )

    try:
        df = (
            Simulator(
                model,
                transform=[getattr(model, name) for name in settings.variables],
            )
            .solve(
                save_at=save_at,
                solver=LSODA(
                    atol=settings.absolute,
                    rtol=settings.relative,
                ),
            )
            .reset_index()
        )
    except TypeError as e:
        if Unsupported.__name__ in e.args[0]:
            xfail("unsupported symbolite operation")
        raise e
    except KeyError:
        xfail("dependency solving error")

    df_expected = loader.read_results()
    assert_frame_equal(
        df[settings.variables],
        df_expected[settings.variables],
        atol=settings.absolute * ATOL_FACTOR.get(model_id, 1),
        rtol=settings.relative,
        check_dtype=False,
    )



import math

from pytest import mark, param, raises

from ...core import Parameter, Species
from . import types
from .importer import convert, nan_to_none


@mark.parametrize("value", [None, math.nan, 1.0])
@mark.parametrize("units", [None])
@mark.parametrize("constant", [False, True])
def test_parameter(value, units, constant):
    name = types.ID("p")
    model = types.Model(
        parameters=[
            types.Parameter(
                id=name,
                value=value,
                units=units,
                constant=constant,
            )
        ],
    )

    compartment = convert(model, name="model")
    p = getattr(compartment, name)
    assert isinstance(p, Parameter)
    assert p.default == nan_to_none(value)


@mark.parametrize("initial_amount", [None, math.nan, 0, 1])
@mark.parametrize("initial_concentration", [None, math.nan, 0, 1])
@mark.parametrize("substance_units", [None])
@mark.parametrize("has_only_substance_units", [False, True])
@mark.parametrize("boundary_condition", [False, True])
@mark.parametrize("constant", [False, True])
@mark.parametrize("conversion_factor", [None])
def test_species(
    initial_amount,
    initial_concentration,
    substance_units,
    has_only_substance_units,
    boundary_condition,
    constant,
    conversion_factor,
):
    name = types.ID("s")
    compartment = types.ID("c")
    model = types.Model(
        species=[
            types.Species(
                id=name,
                compartment=compartment,
                initial_amount=initial_amount,
                initial_concentration=initial_concentration,
                substance_units=substance_units,
                has_only_substance_units=has_only_substance_units,
                boundary_condition=boundary_condition,
                constant=constant,
                conversion_factor=conversion_factor,
            )
        ],
    )

    initial_amount = nan_to_none(initial_amount)
    initial_concentration = nan_to_none(initial_concentration)
    if initial_amount is not None and initial_concentration is not None:
        with raises(ValueError):
            convert(model, name="model")
    else:
        compartment = convert(model, name="model")
        s = getattr(compartment, name)
        assert isinstance(s, Species)
        if initial_amount is None:
            assert s.initial == initial_concentration
        else:
            assert s.initial == initial_amount


@mark.parametrize("spatial_dimensions", [None, 3])
@mark.parametrize(
    "size",
    [
        param(None, marks=mark.xfail(reason="compartment size must be 1")),
        param(math.nan, marks=mark.xfail(reason="compartment size must be 1")),
        1,
        param(1.5, marks=mark.xfail(reason="compartment size must be 1")),
    ],
)
@mark.parametrize("units", [None])
@mark.parametrize("constant", [False, True])
def test_compartment(spatial_dimensions, size, units, constant):
    name = types.ID("c")
    model = types.Model(
        compartments=[
            types.Compartment(
                id=name,
                spatial_dimensions=spatial_dimensions,
                size=size,
                units=units,
                constant=constant,
            )
        ],
    )

    compartment = convert(model, name="model")
    p = getattr(compartment, name)
    assert isinstance(p, Parameter)
    assert p.default == nan_to_none(size)



from dataclasses import asdict
from functools import singledispatchmethod

import libsbml

from ..mathML import mathMLImporter
from . import types


class Converter:
    def __init__(self):
        self.mathML = mathMLImporter()

    @singledispatchmethod
    def convert(self, x):
        raise NotImplementedError(type(x))

    @convert.register
    def NoneType(self, x: None):
        return None

    @convert.register
    def ListOf(self, x: libsbml.ListOf) -> list:
        return list(map(self.convert, x))

    @convert.register
    def Math(self, x: libsbml.ASTNode):
        return self.mathML.convert(x)

    @convert.register
    def FunctionDefinition(
        self, x: libsbml.FunctionDefinition
    ) -> types.FunctionDefinition:
        # x.isSetMath()
        return types.FunctionDefinition(
            **asdict(self.Base(x)),
            math=self.mathML.compile_function(x.getId(), x.getMath()),
        )

    def Base(self, x: libsbml.SBase) -> types.Base:
        return types.Base(
            id=types.ID(x.getId()) if x.isSetId() else None,
            name=x.getName() if x.isSetName() else None,
            meta_id=x.getMetaId() if x.isSetMetaId() else None,
            sbo_term=x.getSBOTerm() if x.isSetSBOTerm() else None,
            notes=x.getNotesString() if x.isSetNotes() else None,
            annotation=x.getAnnotationString() if x.isSetAnnotation() else None,
        )

    @convert.register
    def Model(self, x: libsbml.Model) -> types.Model:
        return types.Model(
            **asdict(self.Base(x)),
            substance_units=x.getSubstanceUnits() if x.isSetSubstanceUnits() else None,
            time_units=x.getTimeUnits() if x.isSetTimeUnits() else None,
            volume_units=x.getVolumeUnits() if x.isSetVolumeUnits() else None,
            area_units=x.getAreaUnits() if x.isSetAreaUnits() else None,
            length_units=x.getLengthUnits() if x.isSetLengthUnits() else None,
            extent_units=x.getExtentUnits() if x.isSetExtentUnits() else None,
            conversion_factor=x.getConversionFactor()
            if x.isSetConversionFactor()
            else None,
            functions=self.convert(x.getListOfFunctionDefinitions()),
            units=self.convert(x.getListOfUnitDefinitions()),
            compartments=self.convert(x.getListOfCompartments()),
            species=self.convert(x.getListOfSpecies()),
            parameters=self.convert(x.getListOfParameters()),
            initial_assignments=self.convert(x.getListOfInitialAssignments()),
            rules=self.convert(x.getListOfRules()),
            constraints=self.convert(x.getListOfConstraints()),
            reactions=self.convert(x.getListOfReactions()),
            events=self.convert(x.getListOfEvents()),
        )

    @convert.register
    def Parameter(self, x: libsbml.Parameter) -> types.Parameter:
        # x.isSetConstant()
        return types.Parameter(
            **asdict(self.Base(x)),
            value=x.getValue() if x.isSetValue() else None,
            units=x.getUnits() if x.isSetUnits() else None,
            constant=x.getConstant(),
        )

    @convert.register
    def LocalParameter(self, x: libsbml.LocalParameter) -> types.LocalParameter:
        return types.LocalParameter(
            **asdict(self.Base(x)),
            value=x.getValue() if x.isSetValue() else None,
            units=x.getUnits() if x.isSetUnits() else None,
        )

    @convert.register
    def Species(self, x: libsbml.Species) -> types.Species:
        # x.isSetCompartment()
        # x.isSetHasOnlySubstanceUnits()
        # x.isSetBoundaryCondition()
        # x.isSetConstant()
        return types.Species(
            **asdict(self.Base(x)),
            compartment=x.getCompartment(),
            initial_amount=x.getInitialAmount() if x.isSetInitialAmount() else None,
            initial_concentration=x.getInitialConcentration()
            if x.isSetInitialConcentration()
            else None,
            substance_units=x.getSubstanceUnits() if x.isSetSubstanceUnits() else None,
            has_only_substance_units=x.getHasOnlySubstanceUnits(),
            boundary_condition=x.getBoundaryCondition(),
            constant=x.getConstant(),
            conversion_factor=x.getConversionFactor()
            if x.isSetConversionFactor()
            else None,
        )

    @convert.register
    def Compartment(self, x: libsbml.Compartment) -> types.Compartment:
        # x.isSetConstant()
        return types.Compartment(
            **asdict(self.Base(x)),
            spatial_dimensions=x.getSpatialDimensions()
            if x.isSetSpatialDimensions()
            else None,
            size=x.getSize() if x.isSetSize() else None,
            units=x.getUnits() if x.isSetUnits() else None,
            constant=x.getConstant(),
        )

    @convert.register
    def Event(self, x: libsbml.Event) -> types.Event:
        # x.isSetUseValuesFromTriggerTime()
        return types.Event(
            **asdict(self.Base(x)),
            use_values_from_trigger_time=x.getUseValuesFromTriggerTime(),
            trigger=self.convert(x.getTrigger()),
            prority=self.convert(x.getPriority()),
            delay=self.convert(x.getDelay()),
            assignments=self.convert(x.getListOfEventAssignments()),
        )

    @convert.register
    def Priority(self, x: libsbml.Priority) -> types.Priority:
        # x.isSetMath()
        return types.Priority(
            **asdict(self.Base(x)),
            math=self.convert(x.getMath()),
        )

    @convert.register
    def Delay(self, x: libsbml.Delay) -> types.Delay:
        # x.isSetMath()
        return types.Delay(
            **asdict(self.Base(x)),
            math=self.convert(x.getMath()),
        )

    @convert.register
    def EventAssignment(self, x: libsbml.EventAssignment) -> types.EventAssignment:
        # x.isSetMath()
        return types.EventAssignment(
            **asdict(self.Base(x)),
            math=self.convert(x.getMath()),
        )

    @convert.register
    def InitialAssignment(
        self, x: libsbml.InitialAssignment
    ) -> types.InitialAssignment:
        # x.isSetSymbol()
        # x.isSetMath()
        return types.InitialAssignment(
            **asdict(self.Base(x)),
            symbol=x.getSymbol(),
            math=self.convert(x.getMath()),
        )

    @convert.register
    def Rule(self, x: libsbml.Rule) -> types.Rule:
        # x.isSetMath()
        return types.Rule(
            **asdict(self.Base(x)),
            math=self.convert(x.getMath()),
        )

    @convert.register
    def AlgebraicRule(self, x: libsbml.AlgebraicRule) -> types.AlgebraicRule:
        # x.isSetMath()
        return types.AlgebraicRule(
            **asdict(self.Base(x)),
            math=self.convert(x.getMath()),
        )

    @convert.register
    def AssignmentRule(self, x: libsbml.AssignmentRule) -> types.AssignmentRule:
        # x.isSetMath()
        # x.isSetVariable()
        return types.AssignmentRule(
            **asdict(self.Base(x)),
            math=self.convert(x.getMath()),
            variable=x.getVariable(),
        )

    @convert.register
    def RateRule(self, x: libsbml.RateRule) -> types.RateRule:
        # x.isSetMath()
        # x.isSetVariable()
        return types.RateRule(
            **asdict(self.Base(x)),
            math=self.convert(x.getMath()),
            variable=x.getVariable(),
        )

    @convert.register
    def Constraint(self, x: libsbml.Constraint) -> types.Constraint:
        # x.isSetMath()
        # x.isSetMessageString()
        return types.Constraint(
            **asdict(self.Base(x)),
            math=self.convert(x.getMath()),
            message=x.getMessageString(),
        )

    @convert.register
    def Reaction(self, x: libsbml.Reaction) -> types.Reaction:
        # x.isSetReversible()
        return types.Reaction(
            **asdict(self.Base(x)),
            reversible=x.getReversible(),
            fast=x.getFast(),
            compartment=x.getCompartment() if x.isSetCompartment() else None,
            reactants=self.convert(x.getListOfReactants()),
            products=self.convert(x.getListOfProducts()),
            modifiers=self.convert(x.getListOfModifiers()),
            kinetic_law=self.convert(x.getKineticLaw()),
        )

    @convert.register
    def SimpleSpeciesReference(
        self,
        x: libsbml.SimpleSpeciesReference,
    ) -> types.SimpleSpeciesReference:
        return types.SimpleSpeciesReference(
            **asdict(self.Base(x)),
            species=types.ID(x.getSpecies()),
        )

    @convert.register
    def ModifierSpeciesReference(
        self,
        x: libsbml.ModifierSpeciesReference,
    ) -> types.ModifierSpeciesReference:
        return types.ModifierSpeciesReference(
            **asdict(self.Base(x)),
            species=types.ID(x.getSpecies()),
        )

    @convert.register
    def SpeciesReference(self, x: libsbml.SpeciesReference) -> types.SpeciesReference:
        if x.isSetStoichiometry():
            stoichiometry = x.getStoichiometry()
        elif x.isSetStoichiometryMath():
            raise NotImplementedError("math stoichiometry")
        else:
            stoichiometry = None

        # x.isSetConstant()
        return types.SpeciesReference(
            **asdict(self.Base(x)),
            species=types.ID(x.getSpecies()),
            stoichiometry=stoichiometry,
            constant=x.getConstant(),
        )

    @convert.register
    def KineticLaw(self, x: libsbml.KineticLaw) -> types.KineticLaw:
        # x.isSetMath()
        return types.KineticLaw(
            **asdict(self.Base(x)),
            math=self.convert(x.getMath()),
            parameters=self.convert(x.getListOfParameters()),
        )

    @convert.register
    def Trigger(self, x: libsbml.Trigger) -> types.Trigger:
        # x.isSetInitialValue()
        # x.isSetPersistent()
        # x.isSetMath()
        return types.Trigger(
            **asdict(self.Base(x)),
            initial_value=x.getInitialValue(),
            persistent=x.getPersistent(),
            math=self.convert(x.getMath()),
        )

    @convert.register
    def UnitDefinition(self, x: libsbml.UnitDefinition) -> types.UnitDefinition:
        return types.UnitDefinition(
            **asdict(self.Base(x)),
            units=self.convert(x.getListOfUnits()),
        )

    @convert.register
    def Unit(self, x: libsbml.Unit) -> types.Unit:
        # x.isSetKind()
        # x.isSetExponent()
        # x.isSetScale()
        # x.isSetMultiplier()
        return types.Unit(
            kind=unit_kind_map[x.getKind()],
            exponent=x.getExponent(),
            scale=x.getScale(),
            multiplier=x.getMultiplier(),
        )


class _UnitKindMap(dict):
    def __getitem__(self, item) -> types.UnitKind:
        value = super().__getitem__(item)
        if isinstance(value, str):
            raise NotImplementedError(value)
        else:
            return value


unit_kind_map: dict[int, types.UnitKind] = _UnitKindMap(
    {
        libsbml.UNIT_KIND_AMPERE: types.UnitKind.ampere,
        libsbml.UNIT_KIND_COULOMB: types.UnitKind.coulomb,
        libsbml.UNIT_KIND_GRAY: types.UnitKind.gray,
        libsbml.UNIT_KIND_JOULE: types.UnitKind.joule,
        libsbml.UNIT_KIND_LITRE: types.UnitKind.litre,
        libsbml.UNIT_KIND_MOLE: types.UnitKind.mole,
        libsbml.UNIT_KIND_RADIAN: types.UnitKind.radian,
        libsbml.UNIT_KIND_STERADIAN: types.UnitKind.steradian,
        libsbml.UNIT_KIND_WEBER: types.UnitKind.weber,
        libsbml.UNIT_KIND_AVOGADRO: types.UnitKind.avogadro,
        libsbml.UNIT_KIND_DIMENSIONLESS: types.UnitKind.dimensionless,
        libsbml.UNIT_KIND_HENRY: types.UnitKind.henry,
        libsbml.UNIT_KIND_KATAL: types.UnitKind.katal,
        libsbml.UNIT_KIND_LUMEN: types.UnitKind.lumen,
        libsbml.UNIT_KIND_NEWTON: types.UnitKind.newton,
        libsbml.UNIT_KIND_SECOND: types.UnitKind.second,
        libsbml.UNIT_KIND_TESLA: types.UnitKind.tesla,
        libsbml.UNIT_KIND_BECQUEREL: types.UnitKind.becquerel,
        libsbml.UNIT_KIND_FARAD: types.UnitKind.farad,
        libsbml.UNIT_KIND_HERTZ: types.UnitKind.hertz,
        libsbml.UNIT_KIND_KELVIN: types.UnitKind.kelvin,
        libsbml.UNIT_KIND_LUX: types.UnitKind.lux,
        libsbml.UNIT_KIND_OHM: types.UnitKind.ohm,
        libsbml.UNIT_KIND_SIEMENS: types.UnitKind.siemens,
        libsbml.UNIT_KIND_VOLT: types.UnitKind.volt,
        libsbml.UNIT_KIND_CANDELA: types.UnitKind.candela,
        libsbml.UNIT_KIND_GRAM: types.UnitKind.gram,
        libsbml.UNIT_KIND_ITEM: types.UnitKind.item,
        libsbml.UNIT_KIND_KILOGRAM: types.UnitKind.kilogram,
        libsbml.UNIT_KIND_METRE: types.UnitKind.metre,
        libsbml.UNIT_KIND_PASCAL: types.UnitKind.pascal,
        libsbml.UNIT_KIND_SIEVERT: types.UnitKind.sievert,
        libsbml.UNIT_KIND_WATT: types.UnitKind.watt,
        libsbml.UNIT_KIND_LITER: types.UnitKind.litre,
        libsbml.UNIT_KIND_METER: types.UnitKind.metre,
        libsbml.UNIT_KIND_CELSIUS: "UNIT_KIND_CELSIUS",
        libsbml.UNIT_KIND_INVALID: "UNIT_KIND_INVALID",
    }
)



from typing import Sequence

import biomodels

from .sbml import loads


def load(
    model_id: str,
    *,
    name: str | None = None,
    ignore_namespaces: Sequence[str] = [],
):
    omex = biomodels.get_omex(model_id)
    text = omex.master.read_text()
    model = loads(
        text,
        name=name,
        ignore_namespaces=ignore_namespaces,
    )
    return model



from functools import singledispatch as dispatch

import libsbml
from symbolite import Real
from symbolite.abstract import real
from symbolite.core.symbolite_object import get_symbolite_info

mapper = {
    real.add: libsbml.AST_PLUS,
    real.sub: libsbml.AST_MINUS,
    real.neg: libsbml.AST_MINUS,
    real.mul: libsbml.AST_TIMES,
    real.truediv: libsbml.AST_DIVIDE,
    real.pow: libsbml.AST_POWER,
    "libsbml.AST_INTEGER": libsbml.AST_INTEGER,
    "libsbml.AST_REAL": libsbml.AST_REAL,
    "libsbml.AST_REAL_E": libsbml.AST_REAL_E,
    "AST_RATIONAL": libsbml.AST_RATIONAL,
    "libsbml.AST_NAME": libsbml.AST_NAME,
    "AST_NAME_AVOGADRO": libsbml.AST_NAME_AVOGADRO,
    "libsbml.AST_NAME_TIME": libsbml.AST_NAME_TIME,
    real.e: libsbml.AST_CONSTANT_E,
    False: libsbml.AST_CONSTANT_FALSE,
    real.pi: libsbml.AST_CONSTANT_PI,
    True: libsbml.AST_CONSTANT_TRUE,
    "AST_LAMBDA": libsbml.AST_LAMBDA,
    "AST_FUNCTION": libsbml.AST_FUNCTION,
    real.abs: libsbml.AST_FUNCTION_ABS,
    real.acos: libsbml.AST_FUNCTION_ARCCOS,
    real.acosh: libsbml.AST_FUNCTION_ARCCOSH,
    "AST_FUNCTION_ARCCOT": libsbml.AST_FUNCTION_ARCCOT,
    "AST_FUNCTION_ARCCOTH": libsbml.AST_FUNCTION_ARCCOTH,
    "AST_FUNCTION_ARCCSC": libsbml.AST_FUNCTION_ARCCSC,
    "AST_FUNCTION_ARCCSCH": libsbml.AST_FUNCTION_ARCCSCH,
    "AST_FUNCTION_ARCSEC": libsbml.AST_FUNCTION_ARCSEC,
    "AST_FUNCTION_ARCSECH": libsbml.AST_FUNCTION_ARCSECH,
    real.asin: libsbml.AST_FUNCTION_ARCSIN,
    real.asinh: libsbml.AST_FUNCTION_ARCSINH,
    real.atan: libsbml.AST_FUNCTION_ARCTAN,
    real.atanh: libsbml.AST_FUNCTION_ARCTANH,
    real.ceil: libsbml.AST_FUNCTION_CEILING,
    real.cos: libsbml.AST_FUNCTION_COS,
    real.cosh: libsbml.AST_FUNCTION_COSH,
    "AST_FUNCTION_COT": libsbml.AST_FUNCTION_COT,
    "AST_FUNCTION_COTH": libsbml.AST_FUNCTION_COTH,
    "AST_FUNCTION_CSC": libsbml.AST_FUNCTION_CSC,
    "AST_FUNCTION_CSCH": libsbml.AST_FUNCTION_CSCH,
    "AST_FUNCTION_DELAY": libsbml.AST_FUNCTION_DELAY,
    real.exp: libsbml.AST_FUNCTION_EXP,
    real.factorial: libsbml.AST_FUNCTION_FACTORIAL,
    real.floor: libsbml.AST_FUNCTION_FLOOR,
    real.log: libsbml.AST_FUNCTION_LN,
    real.log10: libsbml.AST_FUNCTION_LOG,
    "AST_FUNCTION_PIECEWISE": libsbml.AST_FUNCTION_PIECEWISE,
    real.pow: libsbml.AST_FUNCTION_POWER,
    real.sqrt: libsbml.AST_FUNCTION_ROOT,
    "AST_FUNCTION_SEC": libsbml.AST_FUNCTION_SEC,
    "AST_FUNCTION_SECH": libsbml.AST_FUNCTION_SECH,
    real.sin: libsbml.AST_FUNCTION_SIN,
    real.sinh: libsbml.AST_FUNCTION_SINH,
    real.tan: libsbml.AST_FUNCTION_TAN,
    real.tanh: libsbml.AST_FUNCTION_TANH,
    real.and_: libsbml.AST_LOGICAL_AND,
    real.invert: libsbml.AST_LOGICAL_NOT,
    real.or_: libsbml.AST_LOGICAL_OR,
    real.xor: libsbml.AST_LOGICAL_XOR,
    real.eq: libsbml.AST_RELATIONAL_EQ,
    real.ge: libsbml.AST_RELATIONAL_GEQ,
    real.gt: libsbml.AST_RELATIONAL_GT,
    real.le: libsbml.AST_RELATIONAL_LEQ,
    real.lt: libsbml.AST_RELATIONAL_LT,
    real.ne: libsbml.AST_RELATIONAL_NEQ,
    "AST_END_OF_CORE": libsbml.AST_END_OF_CORE,
    "AST_FUNCTION_MAX": libsbml.AST_FUNCTION_MAX,
    "AST_FUNCTION_MIN": libsbml.AST_FUNCTION_MIN,
    "AST_FUNCTION_QUOTIENT": libsbml.AST_FUNCTION_QUOTIENT,
    "AST_FUNCTION_RATE_OF": libsbml.AST_FUNCTION_RATE_OF,
    "AST_FUNCTION_REM": libsbml.AST_FUNCTION_REM,
    "AST_LOGICAL_IMPLIES": libsbml.AST_LOGICAL_IMPLIES,
    "AST_CSYMBOL_FUNCTION": libsbml.AST_CSYMBOL_FUNCTION,
    "AST_DISTRIB_FUNCTION_NORMAL": libsbml.AST_DISTRIB_FUNCTION_NORMAL,
    "AST_DISTRIB_FUNCTION_UNIFORM": libsbml.AST_DISTRIB_FUNCTION_UNIFORM,
    "AST_DISTRIB_FUNCTION_BERNOULLI": libsbml.AST_DISTRIB_FUNCTION_BERNOULLI,
    "AST_DISTRIB_FUNCTION_BINOMIAL": libsbml.AST_DISTRIB_FUNCTION_BINOMIAL,
    "AST_DISTRIB_FUNCTION_CAUCHY": libsbml.AST_DISTRIB_FUNCTION_CAUCHY,
    "AST_DISTRIB_FUNCTION_CHISQUARE": libsbml.AST_DISTRIB_FUNCTION_CHISQUARE,
    "AST_DISTRIB_FUNCTION_EXPONENTIAL": libsbml.AST_DISTRIB_FUNCTION_EXPONENTIAL,
    "AST_DISTRIB_FUNCTION_GAMMA": libsbml.AST_DISTRIB_FUNCTION_GAMMA,
    "AST_DISTRIB_FUNCTION_LAPLACE": libsbml.AST_DISTRIB_FUNCTION_LAPLACE,
    "AST_DISTRIB_FUNCTION_LOGNORMAL": libsbml.AST_DISTRIB_FUNCTION_LOGNORMAL,
    "AST_DISTRIB_FUNCTION_POISSON": libsbml.AST_DISTRIB_FUNCTION_POISSON,
    "AST_DISTRIB_FUNCTION_RAYLEIGH": libsbml.AST_DISTRIB_FUNCTION_RAYLEIGH,
    "AST_LINEAR_ALGEBRA_VECTOR": libsbml.AST_LINEAR_ALGEBRA_VECTOR,
    "AST_LINEAR_ALGEBRA_SELECTOR": libsbml.AST_LINEAR_ALGEBRA_SELECTOR,
    "AST_LINEAR_ALGEBRA_MATRIX": libsbml.AST_LINEAR_ALGEBRA_MATRIX,
    "AST_LINEAR_ALGEBRA_MATRIXROW": libsbml.AST_LINEAR_ALGEBRA_MATRIXROW,
    "AST_LINEAR_ALGEBRA_DETERMINANT": libsbml.AST_LINEAR_ALGEBRA_DETERMINANT,
    "AST_LINEAR_ALGEBRA_TRANSPOSE": libsbml.AST_LINEAR_ALGEBRA_TRANSPOSE,
    "AST_LINEAR_ALGEBRA_VECTOR_PRODUCT": libsbml.AST_LINEAR_ALGEBRA_VECTOR_PRODUCT,
    "AST_LINEAR_ALGEBRA_SCALAR_PRODUCT": libsbml.AST_LINEAR_ALGEBRA_SCALAR_PRODUCT,
    "AST_LINEAR_ALGEBRA_OUTER_PRODUCT": libsbml.AST_LINEAR_ALGEBRA_OUTER_PRODUCT,
    "AST_LOGICAL_EXISTS": libsbml.AST_LOGICAL_EXISTS,
    "AST_LOGICAL_FORALL": libsbml.AST_LOGICAL_FORALL,
    "AST_STATISTICS_MEAN": libsbml.AST_STATISTICS_MEAN,
    "AST_STATISTICS_MEDIAN": libsbml.AST_STATISTICS_MEDIAN,
    "AST_STATISTICS_MODE": libsbml.AST_STATISTICS_MODE,
    "AST_STATISTICS_MOMENT": libsbml.AST_STATISTICS_MOMENT,
    "AST_STATISTICS_SDEV": libsbml.AST_STATISTICS_SDEV,
    "AST_STATISTICS_VARIANCE": libsbml.AST_STATISTICS_VARIANCE,
    "AST_STATISTICS_MOMENTABOUT": libsbml.AST_STATISTICS_MOMENTABOUT,
    "AST_SERIES_PRODUCT": libsbml.AST_SERIES_PRODUCT,
    "AST_SERIES_SUM": libsbml.AST_SERIES_SUM,
    "AST_SERIES_LIMIT": libsbml.AST_SERIES_LIMIT,
    "AST_SERIES_TENDSTO": libsbml.AST_SERIES_TENDSTO,
    "AST_ALGEBRA_GCD": libsbml.AST_ALGEBRA_GCD,
    "AST_ALGEBRA_CONJUGATE": libsbml.AST_ALGEBRA_CONJUGATE,
    "AST_ALGEBRA_ARG": libsbml.AST_ALGEBRA_ARG,
    "AST_ALGEBRA_REAL": libsbml.AST_ALGEBRA_REAL,
    "AST_ALGEBRA_IMAGINARY": libsbml.AST_ALGEBRA_IMAGINARY,
    "AST_ALGEBRA_LCM": libsbml.AST_ALGEBRA_LCM,
    "AST_RELATIONS_EQUIVALENT": libsbml.AST_RELATIONS_EQUIVALENT,
    "AST_RELATIONS_APPROX": libsbml.AST_RELATIONS_APPROX,
    "AST_RELATIONS_FACTOROF": libsbml.AST_RELATIONS_FACTOROF,
    "AST_CALCULUS_INT": libsbml.AST_CALCULUS_INT,
    "AST_CALCULUS_DIFF": libsbml.AST_CALCULUS_DIFF,
    "AST_CALCULUS_PARTIALDIFF": libsbml.AST_CALCULUS_PARTIALDIFF,
    "AST_CALCULUS_LOWLIMIT": libsbml.AST_CALCULUS_LOWLIMIT,
    "AST_CALCULUS_UPLIMIT": libsbml.AST_CALCULUS_UPLIMIT,
    "AST_CALCULUS_DIVERGENCE": libsbml.AST_CALCULUS_DIVERGENCE,
    "AST_CALCULUS_GRAD": libsbml.AST_CALCULUS_GRAD,
    "AST_CALCULUS_CURL": libsbml.AST_CALCULUS_CURL,
    "AST_CALCULUS_LAPLACIAN": libsbml.AST_CALCULUS_LAPLACIAN,
    "AST_SET_THEORY_SET": libsbml.AST_SET_THEORY_SET,
    "AST_SET_THEORY_LIST": libsbml.AST_SET_THEORY_LIST,
    "AST_SET_THEORY_UNION": libsbml.AST_SET_THEORY_UNION,
    "AST_SET_THEORY_INTERSECT": libsbml.AST_SET_THEORY_INTERSECT,
    "AST_SET_THEORY_IN": libsbml.AST_SET_THEORY_IN,
    "AST_SET_THEORY_NOTIN": libsbml.AST_SET_THEORY_NOTIN,
    "AST_SET_THEORY_SUBSET": libsbml.AST_SET_THEORY_SUBSET,
    "AST_SET_THEORY_PRSUBSET": libsbml.AST_SET_THEORY_PRSUBSET,
    "AST_SET_THEORY_NOTSUBSET": libsbml.AST_SET_THEORY_NOTSUBSET,
    "AST_SET_THEORY_NOTPRSUBSET": libsbml.AST_SET_THEORY_NOTPRSUBSET,
    "AST_SET_THEORY_SETDIFF": libsbml.AST_SET_THEORY_SETDIFF,
    "AST_SET_THEORY_CARD": libsbml.AST_SET_THEORY_CARD,
    "AST_SET_THEORY_CARTESIANPRODUCT": libsbml.AST_SET_THEORY_CARTESIANPRODUCT,
    "AST_CONSTANT_IMAGINARYI": libsbml.AST_CONSTANT_IMAGINARYI,
    "AST_CONSTANTS_INTEGERS": libsbml.AST_CONSTANTS_INTEGERS,
    "AST_CONSTANTS_REALS": libsbml.AST_CONSTANTS_REALS,
    "AST_CONSTANTS_RATIONALS": libsbml.AST_CONSTANTS_RATIONALS,
    "AST_CONSTANTS_NATURALNUMBERS": libsbml.AST_CONSTANTS_NATURALNUMBERS,
    "AST_CONSTANTS_COMPLEXES": libsbml.AST_CONSTANTS_COMPLEXES,
    "AST_CONSTANTS_PRIMES": libsbml.AST_CONSTANTS_PRIMES,
    "AST_CONSTANTS_EMPTYSET": libsbml.AST_CONSTANTS_EMPTYSET,
    "AST_CONSTANTS_EULERGAMMA": libsbml.AST_CONSTANTS_EULERGAMMA,
    "AST_BASIC_CONTENT_INTERVAL": libsbml.AST_BASIC_CONTENT_INTERVAL,
    "AST_BASIC_CONTENT_INVERSE": libsbml.AST_BASIC_CONTENT_INVERSE,
    "AST_BASIC_CONTENT_CONDITION": libsbml.AST_BASIC_CONTENT_CONDITION,
    "AST_BASIC_CONTENT_DECLARE": libsbml.AST_BASIC_CONTENT_DECLARE,
    "AST_BASIC_CONTENT_COMPOSE": libsbml.AST_BASIC_CONTENT_COMPOSE,
    "AST_BASIC_CONTENT_IDENT": libsbml.AST_BASIC_CONTENT_IDENT,
    "AST_BASIC_CONTENT_DOMAIN": libsbml.AST_BASIC_CONTENT_DOMAIN,
    "AST_BASIC_CONTENT_CODOMAIN": libsbml.AST_BASIC_CONTENT_CODOMAIN,
    "AST_BASIC_CONTENT_IMAGE": libsbml.AST_BASIC_CONTENT_IMAGE,
    "AST_BASIC_CONTENT_DOMAINOFAPPLICATION": libsbml.AST_BASIC_CONTENT_DOMAINOFAPPLICATION,
    "AST_UNKNOWN": libsbml.AST_UNKNOWN,
}  # type: ignore


@dispatch
def to_mathML(x):
    try:
        return mapper[x]
    except KeyError:
        raise NotImplementedError(f"got type {type(x)}")


@to_mathML.register
def int_to_mathML(x: int):
    node = libsbml.ASTNode(libsbml.AST_INTEGER)
    node.setValue(x)
    return node


@to_mathML.register
def float_to_mathML(x: float):
    node = libsbml.ASTNode(libsbml.AST_REAL)
    node.setValue(x)
    return node


@to_mathML.register
def real_to_mathML(x: Real):
    if x.expression is None:
        node = libsbml.ASTNode(libsbml.AST_NAME)
        node.setId(str(x))
        node.setName(get_symbolite_info(x).value.name)
        return node

    if len(x.expression.kwargs) > 0:
        raise NotImplementedError("mathML does not support functions with kwargs")

    node = libsbml.ASTNode(to_mathML(x.expression.func))
    for arg in x.expression.args:
        node.addChild(to_mathML(arg))
    return node
